/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.pool.impl;

import org.tinygroup.pool.ObjectProvider;
import org.tinygroup.pool.Pool;
import org.tinygroup.pool.PoolMonitor;
import org.tinygroup.pool.PoolState;
import org.tinygroup.pool.exception.PoolException;

//TODO增加使用次数

/**
 * Created by IntelliJ IDEA. User: luoguo Date: 11-3-29 Time: 下午2:21 To change
 * this template use File | Settings | File Templates.
 */
public abstract class AbstractPool<T> implements Pool<T>, PoolMonitor<T> {
	private int initSize = INIT_SIZE;
	private int maxSize = MAX_SIZE;
	private int increaseStep = INCREASE_STEP;
	private boolean testBeforeBorrow;
	private ObjectProvider<T> objectProvider;
	private Object lockObject = new Object();

	private PoolState poolState;
	private int useTimes = MAX_USE_TIMES;
	private String name = null;

	public void initPool(int initSize, int maxSize, int increaseStep,
			boolean testBeforeBorrow, ObjectProvider<T> tObjectProvider) {
		this.initSize = initSize;
		this.maxSize = maxSize;
		this.increaseStep = increaseStep;
		this.testBeforeBorrow = testBeforeBorrow;
		this.objectProvider = tObjectProvider;
		for (int i = 0; i < getInitSize(); i++) {
			T o = getObjectProvider().createObject();
			if (o != null) {
				addObject(new TimesWithObject<T>(o));
			}
		}

	}

	protected abstract void addObject(TimesWithObject<T> timesWithObject);

	protected abstract T borrow() throws PoolException;

	protected abstract void ret(T object);

	protected abstract void stopProvide();

	public void stop() {
		setPoolState(PoolState.STOP);
		synchronized (lockObject) {
			stopProvide();
		}
	}

	public T borrowObject() throws PoolException {
		if (getPoolState() != PoolState.READY.ordinal()) {
			throw new PoolException("Pool state(" + getPoolState()
					+ ") is\'nt ready.");
		}
		synchronized (lockObject) {
			return borrow();
		}
	}

	public void returnObject(T object) {
		synchronized (lockObject) {
			ret(object);
		}
	}

	public int getInitSize() {
		return initSize;
	}

	public void setInitSize(int initSize) {
		this.initSize = initSize;
	}

	public int getMaxSize() {
		return maxSize;
	}

	public void setMaxSize(int maxSize) {
		this.maxSize = maxSize;
	}

	public int getIncreaseStep() {
		return increaseStep;
	}

	public void setIncreaseStep(int increaseStep) {
		this.increaseStep = increaseStep;
	}

	public boolean isTestBeforeBorrow() {
		return testBeforeBorrow;
	}

	public void setTestBeforeBorrow(boolean testBeforeBorrow) {
		this.testBeforeBorrow = testBeforeBorrow;
	}

	public ObjectProvider<T> getObjectProvider() {
		return objectProvider;
	}

	public void setObjectProvider(ObjectProvider<T> objectProvider) {
		this.objectProvider = objectProvider;
	}

	public void setPoolState(PoolState poolState) {
		this.poolState = poolState;
	}

	public void setName(String name) {
		this.name = name;
	}

	public int getUseTimes() {
		return useTimes;
	}

	public void setUseTimes(int useTimes) {
		this.useTimes = useTimes;
	}

	public void start() {
		if (poolState != PoolState.STOP) {
			poolState = PoolState.READY;
		}
	}

	public void pause() {
		poolState = PoolState.PAUSE;
	}

	public int getCurrentPoolSize() {
		return getIdleSize() + getUsingSize();
	}

	public String getName() {
		if (name != null) {
			return name;
		} else {
			return this.getClass().getSimpleName();
		}
	}

	public int getPoolState() {
		return this.poolState.ordinal();
	}

	class TimesWithObject<R> {
		private int times = 0;
		private T object;

		public int getTimes() {
			return times;
		}

		public void setTimes(int times) {
			this.times = times;
		}

		public T getObject() {
			return object;
		}

		TimesWithObject(T object) {
			this.object = object;
		}

		public boolean equals(Object obj) {
			if (obj instanceof TimesWithObject) {
				return this.object.equals(((TimesWithObject<R>) obj).object);
			}
			return object.equals(obj);
		}
	}
}
