/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.pool.impl;

/**
 * TODO 优化动态创建对象，目前是由触发的调用导致的，最终要改成多线程方式解决，
 *      避免调用的人花费时间太多
 */

import java.util.ArrayList;
import java.util.List;

import org.tinygroup.pool.PoolState;
import org.tinygroup.pool.exception.PoolException;

/**
 * Created by IntelliJ IDEA. User: luoguo Date: 11-3-29 Time: 下午2:21 To change
 * this template use File | Settings | File Templates.
 */
public class PoolList<T> extends AbstractPool<T> {
	private List<TimesWithObject<T>> objectIdleList = new ArrayList<TimesWithObject<T>>();
	private List<TimesWithObject<T>> objectUsingList = new ArrayList<TimesWithObject<T>>();

	/**
	 * 当状态是READY时，才可以借出
	 * 
	 * @return
	 * @throws org.snorain.pool.PoolException
	 */
	protected T borrow() throws PoolException {
		// 如果已经用完，且还没到最大可用大小，则增加池大小
		if (objectIdleList.size() == 0 && objectUsingList.size() < getMaxSize()) {
			int size = getMaxSize() - objectUsingList.size() > getIncreaseStep() ? getIncreaseStep()
					: getMaxSize() - objectUsingList.size();
			for (int i = 0; i < size; i++) {
				objectIdleList.add(new TimesWithObject<T>(getObjectProvider()
						.createObject()));
			}
		}
		nextObject: if (objectIdleList.size() > 0) {
			TimesWithObject<T> t = objectIdleList
					.remove(objectIdleList.size() - 1);
			if (isTestBeforeBorrow()) {
				if (getObjectProvider().test(t.getObject())) {
					return getObject(t);
				} else {
					getObjectProvider().releaseObject(t.getObject());
					break nextObject;
					// 测试不通过，则释放对象
					// 继续从空闲链表上取对象
				}
			} else {
				return getObject(t);
			}
		}
		throw new PoolException("Pool is empty.");
	}

	private T getObject(TimesWithObject<T> t) {
		t.setTimes(t.getTimes() + 1);
		objectUsingList.add(t);
		return t.getObject();
	}

	public int getCurrentPoolSize() {
		return objectIdleList.size() + objectUsingList.size();
	}

	/**
	 * 当状态是STOP时，即还即释放
	 * 
	 * @param object
	 */
	protected void ret(T object) {
		if (getPoolState() == PoolState.STOP.ordinal()) {
			objectUsingList.remove(object);
			getObjectProvider().releaseObject(object);
		} else {
			TimesWithObject<T> two = getTimesWithObject(object);
			objectUsingList.remove(two);
			// 如果已经超过最多使用次数，则释放老对象，创建新对象
			if (two.getTimes() == getUseTimes()) {
				getObjectProvider().releaseObject(two.getObject());
				two = new TimesWithObject<T>(getObjectProvider().createObject());
			}
			objectIdleList.add(0, two);

			if (objectIdleList.size() > getInitSize() + getIncreaseStep()) {
				// 如果空闲数已经大于初始数+increaseStep，此时进行回收
				for (int i = 0; i < getIncreaseStep(); i++) {
					TimesWithObject<T> t = objectIdleList.remove(objectIdleList
							.size() - 1);
					getObjectProvider().releaseObject(t.getObject());
				}
			}
		}
	}

	private TimesWithObject<T> getTimesWithObject(T t) {
		for (TimesWithObject<T> two : objectUsingList) {
			if (two.getObject() == t) {
				return two;
			}
		}
		return null;
	}

	public int getIdleSize() {
		return objectIdleList.size();
	}

	public int getUsingSize() {
		return objectUsingList.size();
	}

	/**
	 * 关闭池
	 */
	protected void stopProvide() {
		while (objectIdleList.size() > 0) {
			getObjectProvider().releaseObject(
					objectIdleList.remove(objectIdleList.size() - 1)
							.getObject());
		}
	}

	@Override
	protected void addObject(TimesWithObject<T> timesWithObject) {
		objectIdleList.add(timesWithObject);

	}

}
